"""
whatdido grabs its configuration information from the .didorc file in the root of the git repository.
"""

import os
import json

from typer import Typer, prompt

app = Typer()


class Config:
    """
    The configuration class for the application.
    """

    def __init__(self) -> None:
        self.open_ai_api_key = os.environ.get("OPEN_AI_API_KEY")
        self.open_ai_model = os.environ.get("OPEN_AI_MODEL", "gpt-4-turbo-preview")
        self.temperature = 0.8
        self.instructions = [
            "You're a developer working on a project.",
            "You're about to commit some changes to the project.",
            "Explain the changes. Be as terse as possible. No line breaks, no more than 250 characters.",
        ]


def get_config_dir() -> str:
    return os.path.dirname(get_config_file())


def get_config_file() -> str:
    """
    Get the path to the .env file in the root of the git repository.
    """
    # recurse up from current directory to find the .dido directory
    cwd = os.getcwd()

    while not os.path.exists(os.path.join(cwd, ".dido")) and cwd != "/":
        cwd = os.path.dirname(cwd)

    # if we've reached the root directory, then we can't find the .dido directory
    # so we'll just use the current directory
    if cwd == "/":
        cwd = os.getcwd()

    return os.path.join(cwd, ".dido/config.json")


def validate_config_data(config: dict) -> tuple[bool, list[str]]:
    """
    Validate the configuration.
    """
    if not isinstance(config, dict):
        return (False, ["Configuration must be a dictionary."])

    if "open_ai_api_key" not in config:
        return (False, ["open_ai_api_key is required."])

    if "open_ai_model" not in config:
        return (False, ["open_ai_model is required."])

    return (True, [])


def load_config() -> Config:
    """
    Load configuration from the .env file.
    """
    config = Config()

    config_file = get_config_file()

    # Ensure the file exists
    if not os.path.exists(config_file):
        config = prompt_for_config()
        save_config(config)

    try:
        data = json.load(open(config_file, encoding="utf-8"))
    except json.JSONDecodeError as e:
        raise ValueError(f"Invalid configuration file: {config_file}") from e

    (valid, errors) = validate_config_data(data)

    if not valid:
        raise ValueError(f"Invalid configuration: {errors}")

    config.__dict__.update(data)
    return config


def save_config(config: Config):
    """
    Save configuration to the .env file.
    """
    config_file = get_config_file()

    # ensure folder exists
    os.makedirs(os.path.dirname(config_file), exist_ok=True)

    (valid, errors) = validate_config_data(config.__dict__)

    if not valid:
        raise ValueError(f"Invalid configuration: {errors}")

    with open(config_file, "w", encoding="utf-8") as f:
        json.dump(config.__dict__, f, indent=2)


@app.command("setup")
def prompt_for_config() -> Config:
    """
    Prompt the user for configuration information.
    """
    config = Config()

    config.open_ai_api_key = prompt("OpenAI API Key", default=config.open_ai_api_key)
    config.open_ai_model = prompt("OpenAI Model", default=config.open_ai_model)
    return config


@app.command("read")
def list_config() -> None:
    """
    List the configuration values.
    """
    config = load_config()
    print(json.dumps(config.__dict__, indent=2))


@app.command("get")
def get_config(key: str) -> None:
    """
    Get a configuration value.
    """
    config = load_config()

    if not hasattr(config, key):
        raise ValueError(
            f"Invalid key: {key}. Available keys: {list(config.__dict__.keys())}"
        )

    print(getattr(config, key))


@app.command("set")
def set_config(key: str, value: str) -> None:
    """
    Set a configuration value.
    """
    config = load_config()

    if not hasattr(config, key):
        raise ValueError(
            f"Invalid key: {key}. Available keys: {list(config.__dict__.keys())}"
        )

    setattr(config, key, value)
    save_config(config)


if __name__ == "__main__":
    app()



"""
The CLI entry point for the application.
"""

import re

import openai
import typer

from whatdido.config import load_config, app as config_app
from whatdido.repo import get_repo, assert_in_git_repo
from whatdido.editor import inline_edit

NO_CHANGES_MESSAGE = "No changes to commit. Run `git add` to stage changes."

DIFF_STAGED_SEGMENTS = [
    "HEAD",
    "--staged",
    "--unified=0",
    "--no-color",
    "--no-prefix",
    "--no-renames",
    "--no-ext-diff",
    "--no-textconv",
    "--text",
]
DIFF_CLEANERS = [
    [r"^\d+\s*", ""],  # Remove line numbers,
    [r"\s+", " "],  # Remove extra spaces,
    [r"^\s+", ""],  # Remove leading spaces,
]
MESSAGE_ESCAPERS = [
    ["\n", "\\n"],
    ["\r", "\\r"],
    ["\t", "\\t"],
    ["\v", "\\v"],
    ["\f", "\\f"],
    ['"', '\\"'],
]


config = load_config()

app = typer.Typer()
app.add_typer(config_app, name="config")


def get_terse_git_diff():
    """
    Pares down the git diff to a terse format
    which can be used as input to the OpenAI API.
    Without this function, the diff would be more likely to hit the API's token limit.
    """
    repo = get_repo()

    diff = repo.git.diff(*DIFF_STAGED_SEGMENTS)

    # We don't need all of the spacing, or the line numbers.
    # Use a series of "cleaning" regexes to remove
    # the line numbers and the first 9 spaces.
    for cleaner in DIFF_CLEANERS:
        diff = re.sub(cleaner[0], cleaner[1], diff)

    return diff


def get_ai_summary(diff: str, pst: str):
    """
    Get the AI summary of the changes.
    """

    client = openai.Client(api_key=config.open_ai_api_key)

    messages = [
        {
            "role": "system",
            "content": "\n".join([*config.instructions, pst]),
        }
    ]

    for diff_line in diff.split("\n"):
        messages.append(
            {
                "role": "user",
                "content": diff_line,
            }
        )

    response = client.chat.completions.create(
        model=config.open_ai_model,
        temperature=config.temperature,
        messages=messages,
    )

    message = response.choices[0].message.content
    return message


@app.command("diff")
def run_diff():
    """
    Get the terse diff of the staged changes.
    """
    print(get_terse_git_diff())


def get_summary(pst: str) -> str | None:
    """
    Get the commit message from the user.
    """

    assert_in_git_repo()

    diff = get_terse_git_diff()

    # if there are no changes, there is no need to summarize
    if not diff:
        return None

    commit_summary = get_ai_summary(diff, pst)

    for escaper in MESSAGE_ESCAPERS:
        commit_summary = commit_summary.replace(escaper[0], escaper[1])

    return commit_summary


@app.command("summary")
def run_summary(pst: str = ""):
    """
    Summarize the changes in the repository.
    """
    print(get_summary(pst) or NO_CHANGES_MESSAGE)


@app.command("commit")
def run_commit(pst: str = ""):
    """
    Edit the message using the user's preferred editor.
    """

    commit_summary = get_summary(pst)

    if not commit_summary:
        print(NO_CHANGES_MESSAGE)
        return

    commit_summary = inline_edit(commit_summary)

    repo = get_repo()

    repo.index.commit(commit_summary)


def main():
    """
    Main function to run the app.

    ```bash
    whatdido summarize
    """

    app()



import tempfile
import subprocess


EDITOR_BIN = "vi"


def inline_edit(message: str):
    """Edit the message using vim and return the updated message."""

    outfile = tempfile.NamedTemporaryFile(
        suffix=".txt", mode="w+t", encoding="utf-8", delete=False
    )

    with outfile as f:
        f.write(message)
        f.flush()
        subprocess.run([EDITOR_BIN, f.name], check=True)

        with open(f.name, "r", encoding="utf-8") as f_updated:
            message = f_updated.read()

    return message



"""
Simple wrapper about the GitPython library to interact with a git repository.
"""

import subprocess
from git import Repo


def assert_in_git_repo():
    """
    Ensure we are in a git repository.
    """
    try:
        get_repo()
    except Exception as e:
        raise ValueError("You're not in a git repository.") from e


def get_repo_dir() -> str:
    """
    Get the root directory of the git repository.
    """
    return subprocess.run(
        ["git", "rev-parse", "--show-toplevel"],
        text=True,
        stdout=subprocess.PIPE,
        check=True,
    ).stdout.strip()


def get_repo() -> Repo:
    """
    Get the GitPython Repo object for the current repository.
    @link [GitPython Quick Start Tutorial](https://gitpython.readthedocs.io/en/stable/quickstart.html)
    """
    return Repo.init(get_repo_dir())
